# -*- coding: utf-8 -*-
from .base import *  # pylint: disable=wildcard-import

import os

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-06xo%2f39ohlfrgu&)%1lsrlx6jqy)s%#_-$rc5l_z1+p6w(=r"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["moo.dev.shacklyn.net", "probe.cluster.local"]
CSRF_TRUSTED_ORIGINS = ["https://moo.dev.shacklyn.net", "https://probe.cluster.local"]

STORAGES = {
    "default": {
        "BACKEND": 'storages.backends.s3boto3.S3Boto3Storage',
    },
    "staticfiles": {
        "BACKEND": 'moo.storage.CachedS3Boto3Storage',
    },
}
DEFAULT_FILE_STORAGE = 'storages.backends.s3boto3.S3Boto3Storage'
STATICFILES_STORAGE = 'moo.storage.CachedS3Boto3Storage'
AWS_STORAGE_BUCKET_NAME = 'django-moo-dev-assets-386413725601-us-east-2'
AWS_S3_ADDRESSING_STYLE = "virtual"

STATIC_URL = f"https://{AWS_STORAGE_BUCKET_NAME}.s3.amazonaws.com/"
STATIC_ROOT = '/usr/src/app/static'

COMPRESS_ENABLED = True
COMPRESS_ROOT = STATIC_ROOT
COMPRESS_STORAGE = STATICFILES_STORAGE
COMPRESS_URL = STATIC_URL

# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "moo",
        "HOST": "bubblehouse-dev-moo-db.cluster-cbepmog7ejmj.us-east-2.rds.amazonaws.com",
        "USER": "moo",
        "PASSWORD": os.getenv("DB_PASSWORD", "moo"),
    }
}

CACHES = {
    'default': {
        'BACKEND': 'redis_cache.RedisCache',
        'LOCATION': 'redis:6379',
    }
}



#!moo verb exitfunc --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is invoked by the LambdaMOO server when an object is moved out of a room, as part of the action of the `move`
primitive. The action defined for the `$room` class is to do nothing.

This verb, and the `enterfunc` verb, can be used for a variety of effects that need to take note of objects moving in
and out of rooms. For example, consider a torch object that casts light on its surroundings. When the torch is moved
out of a room, the light that is casts moves with it. This can be tackled using the room's `exitfunc`, which could check
if the object leaving is a torch, and if it is, the room could become dark. Similarly, when the torch enters a room,
the `enterfunc` can be used to detect this, and brighten the room accordingly.
"""

pass



# Generated by Django 5.1.4 on 2025-04-20 23:04

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0011_alter_object_name_alter_object_unique_name_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='Preposition',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='ArgumentSpecifier',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('prepositionSpecifier', models.CharField(blank=True, choices=[('any', 'any'), ('none', 'none')], db_index=True, max_length=255, null=True)),
                ('objectSpecifier', models.CharField(choices=[('this', 'this'), ('any', 'any'), ('none', 'none')], db_index=True, max_length=255)),
                ('preposition', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='core.preposition')),
            ],
        ),
        migrations.CreateModel(
            name='PrepositionName',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('preposition', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='names', to='core.preposition')),
            ],
        ),
    ]



#!moo verb disfunc --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is invoked when the player disconnects from the LambdaMOO server. The verb defined for the `$player` class
does nothing. You might choose to override this verb to print a special message when you log out, or perhaps to
retrieve some of your possessions.
"""

pass



#!moo verb announce_all_but --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This general purpose verb is used to send a message to everyone in the room except a particular object. This can be
used in situations where we wish to present one view to the world in general, and another view to a particular object,
normally a player. Another common usage is to prevent robots that trigger actions based on a redefined :tell() verb
on themselves from recursing, using something like

place.announce_all_but(this, "message");
"""

skip, *messages = args
for obj in this.contents.all():
    if obj != skip:
        obj.tell(" ".join(messages))



#!moo verb test-nested-verbs test_nested_verbs --on "player class" --dspec none

# pylint: disable=return-outside-function,undefined-variable

from moo.core import context

print(1)
context.caller.invoke_verb("test-nested-verbs-method", 1)



#!moo verb tell --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This verb should be used to send a message to a player. The `$player.tell` filters messages in two different ways, as
shown below. Remember that the player referred to in the code is the player sending the message. `this` refers to the
player receiving the message.

The verb `$player:gag_p` returns `True` if the player sending the message is in the recipient's gag list. For this
verb, the output from any gagged player is ignored, and not printed to the recipient's terminal.

If the `paranoid` level of the recipient is `2`, this means that they wish to see who has sent them a message. The
`$player:whodunnit` verb returns the object reference of the player that sent the message. This is prepended to the
message text, which is then printed to the player.

If the `paranoid` level of the recipient is `1`, then the message and its originator are stored in the property list
`responsible` on the player. The list is kept to `player.lines` length, at most. This option is used for later
processing by the `@check` command.
"""

from moo.core import context

player = context.player
callers = context.caller_stack + [{"caller": player, "verb_name": "", "this": player}]
if not this.gag_p():
    if this.paranoid == 2:
        z = this.whodunnit(callers, [this], [])
        passthrough("(", z.name, " #", z.pk, ") ", *args)
    else:
        passthrough(*args)
        if this.paranoid == 1:
            res = this.responsible + [[callers, args]]
            while len(res) > this.lines:
                res.pop(0)
            this.responsible = res
        else:
            this.responsible = []



import pytest

from .. import tasks
from ..models import Object


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_basic_rollback(t_init: Object, t_wizard: Object):
    tasks.parse_code(
        t_wizard.pk, 'from moo.core import create;create(name="CreatedObjectWithAUniqueName", unique_name=True)'
    )
    with pytest.raises(RuntimeError):
        tasks.parse_code(
            t_wizard.pk,
            """from moo.core import create
o = create(name="ErroredObjectWithAUniqueName", unique_name=True)
raise RuntimeError()
""",
        )
    Object.objects.get(name="CreatedObjectWithAUniqueName")
    with pytest.raises(Object.DoesNotExist):
        Object.objects.get(name="ErroredObjectWithAUniqueName")



#!moo verb gag_p --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This verb returns `True` if the current value of the `player` variable is in the gag list of the object indicated by
the variable `this`, or if a non-player object mentioned in the gag list is in the first elements of the verb's
`context.caller_stack` list.
"""

from moo.core import context

player = context.player
gaglist = this.gaglist

for item in gaglist:
    if item == player:
        return True
    if not item.is_player():
        for frame in context.caller_stack:
            if frame["caller"] == item:
                return True
return False



# -*- coding: utf-8 -*-
"""
Celery Tasks for executing commands or raw Python code.
"""

import logging
import warnings
from typing import Any, Optional

from celery import shared_task
from celery.utils.log import get_task_logger
from django.db import transaction

from moo.core.models import verb

from . import code, exceptions, parse
from .models import Object, Verb

log = get_task_logger(__name__)
background_log = logging.getLogger(f"{__name__}.background")


@shared_task(bind=True)
def parse_command(self, caller_id: int, line: str) -> list[Any]:
    """
    Parse a command-line and invoke the requested verb.

    :param caller_id: the PK of the caller of this command
    :param line: the natural-language command to parse and execute
    :return: a list of output lines
    :raises UserError: if a verb failure happens
    """
    output = []
    task_id = self.request.id
    with transaction.atomic():
        caller = Object.objects.get(pk=caller_id)
        with code.ContextManager(caller, output.append, task_id=task_id) as ctx:
            try:
                log.info(f"{caller}: {line}")
                parse.interpret(ctx, line)
            except exceptions.UserError as e:
                log.error(f"{caller}: {e}")
                output.append(f"[bold red]{e}[/bold red]")
    return output


@shared_task(bind=True)
def parse_code(self, caller_id: int, source: str, runtype: str = "exec") -> list[list[Any], Any]:
    """
    Execute code in a task.

    :param caller_id: the PK of the caller of this command
    :param source: the Python code to execute
    :return: a list of output lines and the result value, if any
    """
    output = []
    task_id = self.request.id
    with transaction.atomic():
        caller = Object.objects.get(pk=caller_id)
        with code.ContextManager(caller, output.append, task_id=task_id):
            result = code.interpret(source, "__main__", runtype=runtype)
    return output, result


@shared_task(bind=True)
def invoke_verb(
    self, *args, caller_id: int = None, this_id: int = None, verb_name: int = None, callback_this_id: Optional[int] = None, callback_verb_name: Optional[int] = None, **kwargs
) -> None:
    """
    Asynchronously execute a Verb, optionally returning the result to another Verb.
    The `print()` method logs to a `moo.core.tasks.background` instead of sending
    to the caller; this could probably be improved.

    :param caller_id: the PK of the caller of this command
    :param verb_id: the PK of the Verb to execute
    :param callback_verb_id: the PK of the verb to send the result to
    """
    from moo.core import context

    task_id = self.request.id
    with transaction.atomic():
        caller = Object.objects.get(pk=caller_id)
        this = Object.objects.get(pk=this_id)
        verb = this.get_verb(verb_name)
        with code.ContextManager(caller, background_log.info, task_id=task_id):
            result = verb(*args, **kwargs)
            if callback_verb_name and callback_this_id:
                invoke_verb.delay(result, caller_id=caller_id, this_id=callback_this_id, verb_name=callback_verb_name)



# -*- coding: utf-8 -*-
from .base import *  # pylint: disable=wildcard-import

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-06xo%2f39ohlfrgu&)%1lsrlx6jqy)s%#_-$rc5l_z1+p6w(=r"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ["localhost"]
CSRF_TRUSTED_ORIGINS = ["https://localhost"]
STATIC_ROOT = "static"

# Database
# https://docs.djangoproject.com/en/3.1/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "moo",
        "HOST": "postgres",
        "USER": "moo",
        "PASSWORD": "moo",
    }
}

# Cache
# https://docs.djangoproject.com/en/4.2/ref/settings/#caches

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.redis.RedisCache",
        "LOCATION": "rediss://redis:6379",
    }
}

# Celery configuration
# https://docs.celeryq.dev/en/stable/userguide/configuration.html

CELERY_BROKER_URL = "redis://redis:6379/0"



#!moo verb i*nventory --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is used to tell a player what s/he has in his/her pockets.
"""

if this.contents.count() > 0:
    this.tell("You are carrying:")
    for thing in this.contents.all():
        this.tell(thing.title())
else:
    this.tell("You are empty-handed.")



#!moo verb moveto --on $thing

# pylint: disable=return-outside-function,undefined-variable

"""
This verb should be used to move an `$thing` object from one location to another. It checks to see that `where` is a
valid object, and that the lock on `where` permits the object to enter. If this is the case, then the
`$root_class.moveto` verb is invoked to actually move the object, using `passthrough()`.
"""

where = args[0]

if where.is_unlocked_for(this):
    return passthrough(where)



# -*- coding: utf-8 -*-
"""
Support utilities for database boostrapping.
"""

import argparse
import importlib.resources
import logging
import shlex
import warnings

from django.conf import settings

log = logging.getLogger(__name__)


class ISpecAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):  # pylint: disable=redefined-outer-name
        """
        Custom action to handle the indirect object specifier.
        """
        values = values or []
        result = {}
        for value in values:
            if ":" not in value:
                raise argparse.ArgumentTypeError(f"Invalid indirect object specifier: {value}")
            preposition, specifier = value.split(":", 1)
            if specifier not in ["this", "any", "none"]:
                raise argparse.ArgumentTypeError(f"Invalid indirect object specifier: {specifier}")
            result[preposition] = specifier
        namespace.ispec = result
        return namespace


parser = argparse.ArgumentParser("moo")
parser.add_argument("subcommand", choices=["verb"])
parser.add_argument("names", nargs="+")
parser.add_argument("--on", help="The object to add or modify the verb on", required=True)
parser.add_argument("--dspec", choices=["this", "any", "none", "either"], default="none", help="The direct object specifier")
parser.add_argument("--ispec", metavar="PREP:SPEC", nargs="+", help="Indirect object specifiers", action=ISpecAction)


def get_source(filename, dataset="default"):
    """
    Get the source code for a verb from a Python package.

    :param filename: The name of the file to get the source code for.
    :type filename: str
    :param dataset: The name of the dataset to get the source code for.
    :type dataset: str
    :return: The source code for the verb.
    :rtype: str
    """
    ref = importlib.resources.files("moo.bootstrap") / f"{dataset}_verbs/{filename}"
    with importlib.resources.as_file(ref) as path:
        with open(path, encoding="utf8") as f:
            return f.read()


def load_python(python_path):
    """
    Execute a provided Python bootstrap file against the provided database.

    :param python_path: The path to the Python file to execute.
    :type python_path: str
    """
    with open(python_path, encoding="utf8") as f:
        src = f.read()
        exec(compile(src, python_path, "exec"), globals(), dict())  # pylint: disable=exec-used


def initialize_dataset(dataset="default"):
    """
    Initialize a new dataset.

    This will create the default objects and permissions for the dataset. Notably, it will
    create a `System Object` that is used to store global properties and verbs.

    It will also create a `Wizard` user that is used to manage the system.

    :param dataset: The name of the dataset to initialize.
    :type dataset: str
    :return: The repository object for the dataset.
    :rtype: Repository
    """
    from moo.core import create, models
    from moo.core.parse import Pattern

    for name in settings.DEFAULT_PERMISSIONS:
        _ = models.Permission.objects.create(name=name)
    Pattern.initializePrepositions()

    repo = models.Repository.objects.get(slug=dataset)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        system = models.Object.objects.create(name="System Object", unique_name=True)
        set_default_permissions = models.Verb.objects.create(
            origin=system,
            repo=repo,
            code=get_source("_system_set_default_permissions.py", dataset=dataset),
        )
        set_default_permissions.names.add(
            models.VerbName.objects.create(verb=set_default_permissions, name="set_default_permissions")
        )
        set_default_permissions.invoked_name = "set_default_permissions"
        set_default_permissions.invoked_object = system
        set_default_permissions(set_default_permissions)
        set_default_permissions(system)
    containers = models.Object.objects.create(name="container class")
    containers.add_verb("accept", code="return True")
    # Create the first real user
    wizard = create(name="Wizard", unique_name=True)
    wizard.add_verb("accept", code="return True")
    wizard.owner = wizard
    wizard.save()
    # Wizard owns containers
    containers.owner = wizard
    containers.save()
    # Wizard owns the system...
    system.owner = wizard
    system.save()
    # ...and the default permissions verb
    set_default_permissions.owner = wizard
    set_default_permissions.save()
    return repo


def load_verb_source(path, system, repo, replace=False):
    from moo.core.models.object import Object
    with open(path, encoding="utf8") as f:
        contents = f.read()
        try:
            first, _ = contents.split("\n", maxsplit=1)
        except ValueError:
            return
        if first.startswith("#!moo "):
            log.debug(f"Loading verb source `{path.name}`...")
            shebang = first[6:]
            try:
                args = parser.parse_args(shlex.split(shebang))
            except SystemExit:
                log.error(f"Invalid shebang in verb source `{path.name}`: {shebang}")
                return
            if args.on.startswith("$"):
                obj = system.get_property(name=args.on[1:])
            else:
                obj = Object.objects.get(name=args.on)
            obj.add_verb(
                *args.names,
                code=contents,
                filename=str(path.resolve()),
                repo=repo,
                direct_object=args.dspec,
                indirect_objects=args.ispec,
                replace=replace,
            )
        else:
            log.debug(f"Skipping verb source `{path.name}`...")


def load_verbs(repo, verb_package):
    """
    Load the verbs from a Python package into the database and associate them with the given repository.

    Verb files should start with a shebang:

    .. code-block::

            #!moo [-h] [--on ON] [--dspec {this,any,none,either}] [--ispec PREP:SPEC [PREP:SPEC ...]] {verb} names [names ...]

            positional arguments:
            {verb}
            names

            options:
            -h, --help            show this help message and exit
            --on ON               The object to add or modify the verb on
            --dspec {this,any,none,either}
                                    The direct object specifier
            --ispec PREP:SPEC [PREP:SPEC ...]
                                    Indirect object specifiers

    :param repo: The repository object for the dataset.
    :type repo: Repository
    :param verb_package: The Python package to load the verbs from.
    :type verb_package: str
    """
    from moo.core.models.object import Object
    system = Object.objects.get(pk=1)

    def _iterate_file_paths(ref):
        if ref.is_dir():
            for subref in ref.iterdir():
                _iterate_file_paths(subref)
        elif ref.is_file():
            with importlib.resources.as_file(ref) as path:
                if str(path).endswith(".py"):
                    load_verb_source(path, system, repo)

    for ref in importlib.resources.files(verb_package).iterdir():
        _iterate_file_paths(ref)



#!moo verb look inspect --on $room --dspec either --ispec at:any

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is a player command used to look at various things in the player's current location. It can be used
meaningfully in other verbs, however, if the verb is called with no arguments. In this case, it calls the `look_self`
on the room which the verb was invoked on. You could use this verb to print the description of a room you are not in,
but it is stylistically better to use the `look_self()` verb of the room.

If a preposition is supplied, via this verb being invoked as a command, and there is no `on` or `in` preposition, then
the verb attempts to match the direct object with something in the room. If the match succeeds, the `look_self` verb of
the matched object is called to tell the player what the object looks like.

If an `on` or `in` preposition was used, then the player wishes to look inside a container of some sort, be it a member of
the `$container` class, or in a player's inventory, for example. An attempt is made to match the indirect object with
something in the room. If it succeeds, then an attempt is made to match the direct object with something inside the
container previously matched. If this final match is made, the `look_self` verb of the matched object is invoked to
print its description.

If the direct object is the empty string, `""`, then the container's `look_self` verb is called to print its description.

Any ambiguous or failed matches produce suitable error messages.
"""

from moo.core import context

if context.parser.has_pobj_str("in"):
    container = context.parser.get_pobj("in")
else:
    container = None

if context.parser.has_dobj() and container is None:
    obj = context.parser.get_dobj()
elif context.parser.has_dobj_str():
    dobj_str = context.parser.get_dobj_str()
    if container:
        qs = container.find(dobj_str)
    else:
        qs = context.player.find(dobj_str) or context.player.location.find(dobj_str)
    if not qs:
        print(f"There is no '{dobj_str}' here.")
        return
    obj = qs[0]
elif context.parser.has_pobj_str("at"):
    pobj_str = context.parser.get_pobj_str("at")
    if container:
        qs = container.find(pobj_str)
    else:
        qs = context.player.find(pobj_str) or context.player.location.find(pobj_str)
    if not qs:
        print(f"There is no '{pobj_str}' here.")
        return
    obj = qs[0]
else:
    obj = context.player.location

obj.look_self()



# -*- coding: utf-8 -*-
"""
Django backend admin support.
"""

from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as BaseUserAdmin
from django.contrib.auth.models import User  # pylint: disable=imported-auth-user

from .forms import VerbAdminForm
from .models import acl, auth, object, property, task, verb  # pylint: disable=redefined-builtin


class VerbInline(admin.TabularInline):
    model = verb.Verb
    fk_name = "origin"
    extra = 1
    exclude = ("code",)
    readonly_fields = ("filename", "direct_object", "owner")


class PropertyInline(admin.TabularInline):
    model = property.Property
    fk_name = "origin"
    extra = 1
    readonly_fields = ("name", "value", "owner")


@admin.register(object.Object)
class ObjectAdmin(admin.ModelAdmin):
    list_display = ("__str__", "unique_name", "owner", "location")
    inlines = [
        VerbInline,
        PropertyInline,
    ]
    raw_id_fields = ("owner", "location")


@admin.register(verb.Verb)
class VerbAdmin(admin.ModelAdmin):
    list_display = ("__str__", "owner", "repo", "ref", "filename")
    form = VerbAdminForm
    raw_id_fields = ("owner", "origin")


@admin.register(verb.VerbName)
class VerbNameAdmin(admin.ModelAdmin):
    raw_id_fields = ("verb",)


@admin.register(verb.Preposition)
class PrepositionAdmin(admin.ModelAdmin):
    list_display = ("prepositions",)

    def prepositions(self, obj):
        return ", ".join([name.name for name in obj.names.all()])


@admin.register(property.Property)
class PropertyAdmin(admin.ModelAdmin):
    raw_id_fields = ("owner", "origin")


admin.register(acl.Permission)


@admin.register(acl.Access)
class AccessAdmin(admin.ModelAdmin):
    list_display = ("rule", "actor", "action", "entity", "origin")
    raw_id_fields = ("object", "verb", "property", "accessor")

    def actor(self, obj):
        return obj.actor()

    def entity(self, obj):
        return obj.entity()

    def origin(self, obj):
        return obj.origin()

    def action(self, obj):
        return obj.permission.name


class PlayerInline(admin.StackedInline):
    model = auth.Player
    can_delete = False


admin.site.unregister(User)


@admin.register(User)
class UserAdmin(BaseUserAdmin):
    inlines = [PlayerInline]


admin.register(task.Task)

@admin.register(auth.Player)
class PlayerAdmin(admin.ModelAdmin):
    model = auth.Player



#!moo verb @gripe --on $player --dspec any

# pylint: disable=return-outside-function,undefined-variable

"""
A gripe is a player complaint or observation that is sent, using MOO mail, to all the administrators in the game. It is
intended to provide a way to report problems with the MOO in a high-priority way that will attract the attention of the
people who can do something about it. A good example of the use of a gripe is to complain about a bug in one of the
core classes, or maybe even a request for something to be added.

The implementation of the gripe concept involves a property on the System Object called `$gripe_recipients`. This is a
list of all the players who will be mailed the `@gripe` message. When a player types in `@gripe` they are taken to the
mail room to enter their message. Any text entered on the line with the `@gripe` command is taken to be the subject of
the gripe message. When the message is finished and sent, it is received by all the people on the `$gripe_recipients`
list.
"""

print("@gripe is not yet implemented.")



"""
Parse command strings sent by the client.

This parser can understand a variety of phrases, but they are all represented
by the (BNF?) form:

<verb>[[[<dobj spec> ]<direct-object> ]+[<prep> [<pobj spec> ]<object-of-the-preposition>]*]

There are a long list of prepositions supported, some of which are interchangeable.
"""

import logging
import re

from django.conf import settings
from django.db.models.query import QuerySet
import more_itertools

from .exceptions import *  # pylint: disable=wildcard-import
from .models import Object, Verb

log = logging.getLogger(__name__)


def interpret(ctx, line):
    """
    For a given user, execute a command.
    """
    from . import context

    lex = Lexer(line)
    parser = Parser(lex, context.player)
    ctx.set_parser(parser)
    try:
        verb = parser.get_verb()
    except Verb.DoesNotExist:
        if context.player.location.has_verb("huh"):
            verb = context.player.location.get_verb("huh")
        else:
            raise
    verb()


def unquote(s):
    return s.strip("'\"").replace("\\'", "'").replace('\\"', '"')


class Pattern:
    PREP_SRC = r"(?:\b)(?P<prep>" + "|".join(sum(settings.PREPOSITIONS, [])) + r")(?:\b)"
    SPEC = r"(?P<spec_str>my|the|a|an|\S+(?:\'s|s\'))"
    PHRASE_SRC = r"(?:" + SPEC + r"\s)?(?P<obj_str>.+)"
    PREP = re.compile(PREP_SRC)
    PHRASE = re.compile(PHRASE_SRC)
    POBJ_TEST = re.compile(PREP_SRC + r"\s" + PHRASE_SRC)
    MULTI_WORD = re.compile(r"((\"|\').+?(?!\\).\2)|(\S+)")

    @classmethod
    def initializePrepositions(cls):
        from .models import Preposition

        for preps in settings.PREPOSITIONS:
            preposition = Preposition.objects.create()
            for name in preps:
                preposition.names.create(name=name)

    def tokenize(self, s):
        """
        Find all words or double-quoted-strings in the text
        """
        iterator = re.finditer(self.MULTI_WORD, s)
        words = []
        qotd_matches = []
        for wordmatch in iterator:
            if wordmatch.group(1):
                qotd_matches.append(wordmatch)
            word = unquote(wordmatch.group())
            words.append(word)
        return words, qotd_matches


class Lexer:
    """
    An instance of this class will identify the various parts of a imperative
    sentence. This may be of use to verb code, as well.
    """

    def __init__(self, command):
        self.command = command

        self.dobj_str = None
        self.dobj_spec_str = None

        pattern = Pattern()
        self.words, qotd_matches = pattern.tokenize(command)

        # Now, find all prepositions
        iterator = re.finditer(pattern.PREP, command)
        prep_matches = []
        for prepmatch in iterator:
            prep_matches.append(prepmatch)

        # this method will be used to filter out prepositions inside quotes
        def nonoverlap(match):
            start, end = match.span()
            for word in qotd_matches:
                word_start, word_end = word.span()
                if word_start <= start < word_end:
                    return False
                elif word_start < end < word_end:
                    return False
            return True

        # nonoverlap() will leave only true non-quoted prepositions
        prep_matches = list(filter(nonoverlap, prep_matches))

        # determine if there is anything after the verb
        if len(self.words) > 1:
            # if there are prepositions, we only look for direct objects
            # until the first preposition
            if prep_matches:
                end = prep_matches[0].start() - 1
            else:
                end = len(command)
            # this is the phrase, which could be [[specifier ]object]
            dobj_phrase = command[len(self.words[0]) + 1 : end]
            match = re.match(pattern.PHRASE, dobj_phrase)
            if match:
                result = match.groupdict()
                self.dobj_str = unquote(result["obj_str"])
                if result["spec_str"]:
                    self.dobj_spec_str = unquote(result["spec_str"])
                else:
                    self.dobj_spec_str = ""

        self.prepositions = {}
        # iterate through all the prepositional phrase matches
        for index in range(len(prep_matches)):  # pylint: disable=consider-using-enumerate
            start = prep_matches[index].start()
            # if this is the last preposition, then look from here until the end
            if index == len(prep_matches) - 1:
                end = len(command)
            # otherwise, search until the next preposition starts
            else:
                end = prep_matches[index + 1].start() - 1
            prep_phrase = command[start:end]
            phrase_match = re.match(pattern.POBJ_TEST, prep_phrase)
            if not (phrase_match):
                continue

            result = phrase_match.groupdict()

            # if we get a quoted string here, strip the quotes
            result["obj_str"] = unquote(result["obj_str"])

            if result["spec_str"] is None:
                result["spec_str"] = ""

            self.prepositions.setdefault(result["prep"], []).append([result["spec_str"], result["obj_str"], None])


class Parser:  # pylint: disable=too-many-instance-attributes
    """
    The parser instance is created by the avatar. A new instance is created
    for each command invocation.
    """

    def __init__(self, lexer, caller):
        """
        Create a new parser object for the given command, as issued by
        the given caller, using the registry.
        """
        self.lexer = lexer
        self.caller = caller

        self.this = None
        self.verb = None
        self.prepositions = lexer.prepositions
        self.command = lexer.command
        self.words = lexer.words
        self.dobj_str = lexer.dobj_str
        self.dobj_spec_str = lexer.dobj_spec_str

        if self.lexer:
            for matches in self.prepositions.values():
                for record in matches:
                    spec, name, _ = record
                    # look for an object with this name/specifier
                    obj = self.find_object(spec, name)
                    # try again (maybe it just looked like a specifier)
                    if not obj and spec:
                        name = f"{spec} {name}"
                        spec = ""
                        obj = self.find_object(spec, name)
                    # one last shot for pronouns
                    if not (obj):
                        obj = self.get_pronoun_object(name)
                    record[2] = obj
        if self.dobj_str:
            # look for an object with this name/specifier
            self.dobj = self.find_object(self.dobj_spec_str, self.dobj_str)
            # try again (maybe it just looked like a specifier)
            if not self.dobj and self.dobj_spec_str:
                dobj = self.find_object(None, self.dobj_spec_str + " " + self.dobj_str)
                if dobj:
                    # if we found an object, then correct the saved strings
                    self.dobj_str = self.dobj_spec_str + " " + self.dobj_str
                    self.dobj_spec_str = ""
                    self.dobj = dobj
            # if there's nothing with this name, then we look for
            # pronouns before giving up
            if not (self.dobj):
                self.dobj = self.get_pronoun_object(self.dobj_str)
        else:
            # didn't find anything, probably because nothing was there.
            self.dobj = None
            self.dobj_str = None

    def find_object(self, specifier, name, return_list=False):
        """
        Look for an object, with the optional specifier, in the area
        around the person who entered this command. If the posessive
        form is used (i.e., "Bill's spoon") and that person is not
        here, a NoSuchObjectError is thrown for that person.
        """
        result = None
        search = None

        if specifier == "my":
            search = self.caller
        elif specifier and specifier.find("'") != -1:
            person = specifier[0 : specifier.index("'")]
            location = self.caller.location
            if location:
                search = location.find(person)
        else:
            search = self.caller

        if isinstance(search, QuerySet):
            if len(search) > 1:
                raise AmbiguousObjectError(name, search)
            search = search[0]
        if name and search:
            result = search.find(name)
            if not result:
                result = self.caller.location.find(name)

        if len(result) == 1:
            return result[0]
        elif return_list:
            return result
        elif not result:
            return None
        raise AmbiguousObjectError(name, result)

    def get_verb(self):
        """
        For each of these items the parser will look for a verb using the
        following rules:

            1. For each object in the search order, look for a verb with the
               same name as the first word in the command by traversing its
               inheritance chain.
            2. For each matching verb found, check if the verb's direct object
               and indirect object specifiers match the direct object and the
               objects of the prepositions already defined in the parser.
            3. Continue until a verb is found or the end of the chain is reached.
            4. Verbs found later in the search order will take precedence
               over verbs found earlier in the search order.
        """
        if self.verb:
            return self.verb

        search_order = filter(None, more_itertools.collapse([
            self.caller,
            list(self.caller.contents.all()),
            self.caller.location,
            self.dobj,
            [[x[2] for x in prep] for prep in self.prepositions.values()]
        ]))

        for obj in search_order:
            try:
                if verb := obj.parse_verb(self):
                    self.this = obj
                    self.verb = verb
            except Verb.DoesNotExist:
                continue

        if not (self.this):
            raise Verb.DoesNotExist("parser: " + self.words[0])

        self.verb.invoked_name = self.words[0]
        self.verb.invoked_object = self.this
        return self.verb

    def get_pronoun_object(self, pronoun):
        """
        Also, a object number (starting with a #) will
        return the object for that id.
        """
        if pronoun == "me":
            return self.caller
        elif pronoun == "here":
            return self.caller.location
        elif pronoun[0] == "#":
            return Object.objects.get(pk=int(pronoun[1:]))
        else:
            return None

    def get_dobj(self, lookup=False):
        """
        Get the direct object for this parser. If there was no
        direct object found, raise a NoSuchObjectError
        """
        if lookup and self.dobj_str is not None:
            from moo.core import lookup
            return lookup(self.dobj_str)
        if not (self.dobj):
            raise Object.DoesNotExist(self.dobj_str)
        return self.dobj

    def get_pobj(self, *preps, lookup=False):
        """
        Get the object(s) for the given preposition(s). If there was no
        object found, raise a Object.DoesNotExist; if the preposition
        was not found, raise a NoSuchPrepositionError.
        """
        matches = []
        for prep in preps:
            if not (prep in self.prepositions):
                raise NoSuchPrepositionError(prep)
            for item in self.prepositions[prep]:
                if lookup and item[1] is not None:
                    from moo.core import lookup
                    matches.append(lookup(item[1]))
                elif item[2]:
                    matches.append(item[2])
        if len(matches) > 1:
            raise AmbiguousObjectError(','.join(preps), matches)
        if not (matches):
            raise Object.DoesNotExist(self.prepositions[prep][0][1])
        return matches[0]

    def get_dobj_str(self):
        """
        Get the direct object **string** for this parser. If there was no
        direct object **string** found, raise a NoSuchObjectError
        """
        if not (self.dobj_str):
            raise Object.DoesNotExist("direct object")
        return self.dobj_str

    def get_pobj_str(self, prep, return_list=False):
        """
        Get the object **string** for the given preposition. If there was no
        object **string** found, raise a NoSuchObjectError; if the preposition
        was not found, raise a NoSuchPrepositionError.
        """
        if not (prep in self.prepositions):
            raise NoSuchPrepositionError(prep)
        matches = []
        for item in self.prepositions[prep]:
            if item[1]:
                matches.append(item[1])
        if len(matches) > 1:
            if return_list:
                return matches
            else:
                raise matches[0]
        elif not (matches):
            raise Object.DoesNotExist(self.prepositions[prep][0][1])
        return self.prepositions[prep][0][1]

    def get_pobj_spec_str(self, prep, return_list=False):
        """
        Get the object **specifier** for the given preposition. If there was no
        object **specifier** found, return the empty string; if the preposition
        was not found, raise a NoSuchPrepositionError.
        """
        if not (prep in self.prepositions):
            raise NoSuchPrepositionError(prep)
        matches = []
        for item in self.prepositions[prep]:
            matches.append(item[0])
        if len(matches) > 1:
            if return_list:
                return matches
            else:
                return matches[0]
        return self.prepositions[prep][0][0]

    def has_dobj(self):
        """
        Was a direct object found?
        """
        return self.dobj is not None

    def has_pobj(self, *preps):
        """
        Was an object for this preposition found?
        """
        found_prep = False
        for prep in preps:
            if prep not in self.prepositions:
                return False
            for item in self.prepositions[prep]:
                if item[2]:
                    found_prep = True
                    break
            else:
                continue
            break
        return found_prep

    def has_dobj_str(self):
        """
        Was a direct object string found?
        """
        return self.dobj_str is not None

    def has_pobj_str(self, prep):
        """
        Was a object string for this preposition found?
        """
        if prep not in self.prepositions:
            return False

        found_prep = False

        for item in self.prepositions[prep]:
            if item[1]:
                found_prep = True
                break
        return found_prep



# -*- coding: utf-8 -*-
"""
Prompt-Toolkit interface
"""

import asyncio
import logging
import pickle

from asgiref.sync import sync_to_async
from kombu import Exchange, Queue, simple
from prompt_toolkit import ANSI
from prompt_toolkit.application import run_in_terminal
from prompt_toolkit.shortcuts import print_formatted_text
from prompt_toolkit.shortcuts.prompt import PromptSession
from prompt_toolkit.styles import Style
from rich.console import Console

from ..celery import app
from ..core import models, tasks

log = logging.getLogger(__name__)


async def embed(
    user: models.User,
) -> None:
    repl = MooPrompt(user)
    await asyncio.wait([asyncio.ensure_future(f()) for f in (repl.process_commands, repl.process_messages)])


class MooPrompt:
    style = Style.from_dict(
        {
            # User input (default text).
            "": "#ffffff",
            # Prompt.
            "name": "#884444",
            "at": "#00aa00",
            "colon": "#0000aa",
            "pound": "#00aa00",
            "location": "#00aa55",
        }
    )

    def __init__(self, user, *a, **kw):
        self.user = user
        self.is_exiting = False

    async def process_commands(self):
        prompt_session = PromptSession()
        try:
            while not self.is_exiting:
                message = await self.generate_prompt()
                line = await prompt_session.prompt_async(message, style=self.style)
                await self.handle_command(line)
        except EOFError:
            self.is_exiting = True
        except KeyboardInterrupt:
            self.is_exiting = True
        except:  # pylint: disable=bare-except
            log.exception("Error in command processing")
        log.debug("REPL is exiting, stopping main thread...")

    @sync_to_async
    def generate_prompt(self):
        caller = self.user.player.avatar
        caller.refresh_from_db()
        return [
            ("class:name", str(caller.name)),
            ("class:at", "@"),
            ("class:location", str(caller.location.name)),
            ("class:colon", ":"),
            ("class:pound", "$ "),
        ]

    @sync_to_async
    def handle_command(self, line: str) -> object:
        """
        Parse the command and execute it.
        """
        caller = self.user.player.avatar
        log.info(f"{caller}: {line}")
        ct = tasks.parse_command.delay(caller.pk, line)
        try:
            output = ct.get()
            for item in output:
                self.writer(item)
        except:  # pylint: disable=bare-except
            import traceback

            self.writer(f"[bold red]{traceback.format_exc()}[/bold red]")

    def writer(self, s, is_error=False):
        console = Console(color_system="truecolor")
        with console.capture() as capture:
            console.print(s)
        content = capture.get()
        print_formatted_text(ANSI(content))

    async def process_messages(self) -> None:
        await asyncio.sleep(1)
        try:
            with app.default_connection() as conn:
                channel = conn.channel()
                queue = Queue(
                    "messages", Exchange("moo", type="direct", channel=channel), f"user-{self.user.pk}", channel=channel
                )
                while not self.is_exiting:
                    sb = simple.SimpleBuffer(channel, queue, no_ack=True)
                    try:
                        msg = sb.get_nowait()
                    except sb.Empty:
                        await asyncio.sleep(1)
                        continue
                    if msg:
                        content = pickle.loads(msg.body)
                        await run_in_terminal(lambda: self.writer(content["message"]))
                    sb.close()
        except:  # pylint: disable=bare-except
            log.exception("Stopping message processing")
        log.debug("REPL is exiting, stopping messages thread...")



#!moo verb test-async-verbs --on "player class" --dspec any

# pylint: disable=return-outside-function,undefined-variable

from moo.core import context, invoke

counter = 1
if args and len(args):  # pylint: disable=undefined-variable  # type: ignore
    counter = args[0] + 1  # pylint: disable=undefined-variable  # type: ignore

print(counter)

if counter < 10:
    verb = context.caller.get_verb("test-async-verbs")
    invoke(counter, verb=verb)



# -*- coding: utf-8 -*-
"""
Custom Django forms used in Admin interface.
"""

from django import forms
from django_ace import AceWidget

from .models import verb


class VerbAdminForm(forms.ModelForm):
    class Meta:
        model = verb.Verb
        fields = "__all__"
        widgets = {
            "code": AceWidget(
                mode="python",
                theme="solarized_dark",
                wordwrap=False,
                width="1000px",
                height="600px",
                minlines=None,
                maxlines=None,
                showprintmargin=False,
                showinvisibles=False,
                usesofttabs=True,
                tabsize=4,
                fontsize=None,
                toolbar=True,
                readonly=False,
                showgutter=True,
                behaviours=True,
            )
        }



#!moo verb is_unlocked_for --on $root_class

# pylint: disable=return-outside-function,undefined-variable

"""
Returns `True` if the object is unlocked for the argument. If the value of this.key is None, the object is unlocked. If
this is not the case. the verb $lock_utils:eval_key() is used to determine the result.
"""

thing = args[0]

if not this.key:
    return True

return _.lock_utils.eval_key(this.key, thing)



# Generated by Django 4.2.5 on 2023-11-27 02:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0002_alter_object_observers_alter_object_unique_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="verb",
            name="ability",
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="verb",
            name="method",
            field=models.BooleanField(default=False),
        ),
    ]



#!moo verb is_open is_locked --on $exit --dspec this

# pylint: disable=return-outside-function,undefined-variable

door = this

if verb_name == "is_open":
    prop_name = "open"
elif verb_name == "is_locked":
    prop_name = "locked"
else:
    print("Unknown verb name for door state check: %s" % verb_name)
    return False

if door.has_property(prop_name) and door.get_property(prop_name):
    return True
else:
    return False
